<h3>Navigation</h3>
{{ toc }}
{% if theme_extra_nav_links %}
<hr />
<ul>
    {% for text, uri in theme_extra_nav_links.items() %}
    <li class="toc-l1"><a href="{{ uri }}">{{ text }}</a></li>
    {% endfor %}
</ul>
{% endif %}
